package exchanger

import (
	"context"
	"errors"
	"fmt"
	"sync"
	"time"

	"github.com/Rican7/retry"
	"github.com/Rican7/retry/backoff"
	"github.com/Rican7/retry/strategy"
	"github.com/link33/sidecar/api"
	"github.com/link33/sidecar/internal/checker"
	"github.com/link33/sidecar/internal/peermgr"
	"github.com/link33/sidecar/internal/repo"
	"github.com/link33/sidecar/internal/router"
	"github.com/link33/sidecar/internal/syncer"
	"github.com/link33/sidecar/model/pb"
	"github.com/meshplus/bitxhub-kit/storage"
	"github.com/sirupsen/logrus"
	"go.uber.org/atomic"
)

type Exchanger struct {
	mode        string
	appchainDID string
	store       storage.Storage
	syncer      syncer.Syncer // WrapperSyncer represents the necessary data for sync tx wrappers from bitxhub
	router      router.Router // 占时不使用

	apiServer       *api.Server //直连模式使用
	peerMgr         peermgr.PeerManager
	checker         checker.Checker
	sendIBTPCounter atomic.Uint64
	sendIBTPTimer   atomic.Duration
	ch              chan struct{} //control the concurrent count
	ibtps           sync.Map
	receipts        sync.Map

	logger logrus.FieldLogger
	ctx    context.Context
	cancel context.CancelFunc
}

func New(typ, appchainDID string, meta *pb.Interchain, opts ...Option) (*Exchanger, error) {
	config := GenerateConfig(opts...)

	ctx, cancel := context.WithCancel(context.Background())
	return &Exchanger{
		checker:     config.checker,
		apiServer:   config.apiServer,
		peerMgr:     config.peerMgr,
		syncer:      config.syncer,
		store:       config.store,
		router:      config.router,
		logger:      config.logger,
		ch:          make(chan struct{}, 100),
		mode:        typ,
		appchainDID: appchainDID,
		ctx:         ctx,
		cancel:      cancel,
	}, nil
}

func (ex *Exchanger) Start() error {
	ex.logger.Info("Exchanger started")
	return nil
}

//直链模式
func (ex *Exchanger) startWithDirectMode() error {
	if err := ex.apiServer.Start(); err != nil {
		return fmt.Errorf("peerMgr start: %w", err)
	}

	if err := ex.peerMgr.RegisterConnectHandler(ex.handleNewConnection); err != nil {
		return fmt.Errorf("register on connection handler: %w", err)
	}

	if err := ex.peerMgr.RegisterMsgHandler(pb.Message_INTERCHAIN_META_GET, ex.handleGetInterchainMessage); err != nil {
		return fmt.Errorf("register query interchain msg handler: %w", err)
	}

	// API调用
	if err := ex.peerMgr.RegisterMsgHandler(pb.Message_IBTP_SEND, ex.handleSendIBTPMessage); err != nil {
		return fmt.Errorf("register ibtp handler: %w", err)
	}

	if err := ex.peerMgr.RegisterMsgHandler(pb.Message_IBTP_RECEIPT_SEND, ex.handleSendIBTPReceiptMessage); err != nil {
		return fmt.Errorf("register ibtp handler: %w", err)
	}

	if err := ex.peerMgr.RegisterMsgHandler(pb.Message_IBTP_GET, ex.handleGetIBTPMessage); err != nil {
		return fmt.Errorf("register ibtp receipt handler: %w", err)
	}

	if err := ex.peerMgr.Start(); err != nil {
		return fmt.Errorf("peerMgr start: %w", err)
	}

	go ex.analysisDirectTPS()
	return nil
}

func (ex *Exchanger) startWithRelayMode() error {
	if err := ex.syncer.RegisterRollbackHandler(ex.handleRollback); err != nil {
		return fmt.Errorf("register router handler: %w", err)
	}

	return nil
}

// 中继、直链
func (ex *Exchanger) listenAndSendIBTPFromMnt() {
	ch := ex.mnt.ListenIBTP()
	for {
		select {
		case <-ex.ctx.Done():
			return
		case ibtp, ok := <-ch:
			ex.logger.Info("Receive interchain ibtp from monitor")
			if !ok {
				ex.logger.Warn("Unexpected closed channel while listening on interchain ibtp")
				return
			}
			if err := retry.Retry(func(attempt uint) error {
				if err := ex.sendIBTP(ibtp); err != nil {
					ex.logger.Errorf("Send ibtp: %s", err.Error())
					// if err occurs, try to get new ibtp and resend
					ibtpID := ibtp.ID()
					if err := retry.Retry(func(attempt uint) error {
						ibtp, err = ex.mnt.QueryIBTP(ibtpID)
						if err != nil {
							ex.logger.Errorf("Query ibtp %s from appchain: %s", ibtpID, err.Error())
							return err
						}
						return nil
					}, strategy.Backoff(backoff.Fibonacci(500*time.Millisecond))); err != nil {
						ex.logger.Panic(err)
					}
					return fmt.Errorf("retry sending ibtp")
				}
				return nil
			}, strategy.Backoff(backoff.Fibonacci(500*time.Millisecond))); err != nil {
				ex.logger.Panic(err)
			}
		}
	}
}

// 中继模式
func (ex *Exchanger) listenAndSendIBTPFromSyncer() {
	ch := ex.syncer.ListenIBTP()
	for {
		select {
		case <-ex.ctx.Done():
			return
		case wIbtp, ok := <-ch:
			if !ok {
				ex.logger.Warn("Unexpected closed channel while listening on interchain ibtp")
				return
			}
			entry := ex.logger.WithFields(logrus.Fields{"type": wIbtp.Ibtp.Type, "id": wIbtp.Ibtp.ID()})
			entry.Debugf("Exchanger receives ibtp from syncer")
			switch wIbtp.Ibtp.Type {
			case pb.IBTP_INTERCHAIN:
				ex.applyInterchain(wIbtp, entry) // 发送到HUB上
			case pb.IBTP_RECEIPT_SUCCESS, pb.IBTP_RECEIPT_FAILURE:
				ex.applyReceipt(wIbtp, entry)
			default:
				entry.Errorf("wrong type of ibtp")
			}
		}
	}
}

// 共同
func (ex *Exchanger) sendIBTP(ibtp *pb.IBTP) error {
	entry := ex.logger.WithFields(logrus.Fields{"index": ibtp.Index, "type": ibtp.Type, "to": ibtp.To, "id": ibtp.ID()})
	//TODO!!!! 根据交易决定
	//mode := ibtp.mode
	mode := repo.RelayMode
	switch mode {
	case repo.RelayMode:
		err := ex.syncer.SendIBTP(ibtp)
		if err != nil {
			entry.Errorf("Send ibtp to bitxhub: %s", err.Error())
			return fmt.Errorf("send ibtp to bitxhub: %s", err.Error())
		}
	case repo.DirectMode:
		// send ibtp to another sidecar
		if err := retry.Retry(func(attempt uint) error {
			data, err := ibtp.Marshal()
			if err != nil {
				panic(fmt.Sprintf("marshal ibtp: %s", err.Error()))
			}
			msg := pb.Message(pb.Message_IBTP_SEND, true, data)
			var dst string
			if ibtp.Type == pb.IBTP_INTERCHAIN {
				dst = ibtp.To
			} else {
				dst = ibtp.From
			}

			if err := ex.peerMgr.AsyncSend(dst, msg); err != nil {
				ex.logger.Errorf("Send ibtp to sidecar %s: %s", ibtp.ID(), err.Error())
				return err
			}

			return nil
		}, strategy.Wait(1*time.Second)); err != nil {
			ex.logger.Panic(err)
		}
	}
	entry.Info("Send ibtp success from monitor")
	return nil
}

//主动发起，说明是from自己，
func (ex *Exchanger) queryIBTP(mode string, id, target string) (*pb.IBTP, bool, error) {
	verifiedTx := &pb.VerifiedTx{}
	v := ex.store.Get(pb.IBTPKey(id))
	if v != nil {
		if err := verifiedTx.Unmarshal(v); err != nil {
			return nil, false, err
		}
		return verifiedTx.Tx.GetIBTP(), verifiedTx.Valid, nil
	}

	// query ibtp from counterpart chain
	var (
		ibtp    *pb.IBTP
		isValid bool
		err     error
	)

	// 交易中，如果是用户，根据用户指定，如果没有按默认配置，依次排序。
	switch mode {
	case repo.RelayMode: // 中继模式是查询hub上的ibtp交易根据交易ID。
		ibtp, isValid, err = ex.syncer.QueryIBTP(id)
		if err != nil {
			if errors.Is(err, syncer.ErrIBTPNotFound) {
				ex.logger.Panicf("query ibtp by id %s from bitxhub: %s", id, err.Error())
			}
			return nil, false, fmt.Errorf("query ibtp from bitxhub: %s", err.Error())
		}
	case repo.DirectMode:
		// query ibtp from another sidecar
		msg := pb.Message(pb.Message_IBTP_GET, true, []byte(id))
		result, err := ex.peerMgr.Send(target, msg)
		if err != nil {
			return nil, false, err
		}

		ibtp = &pb.IBTP{}
		if err := ibtp.Unmarshal(result.Payload.Data); err != nil {
			return nil, false, err
		}
	default:
		return nil, false, fmt.Errorf("unsupported sidecar mode")
	}

	return ibtp, isValid, nil
}

func (ex *Exchanger) queryIBTPForRelay(id, target string) (*pb.IBTP, bool, error) {
	verifiedTx := &pb.VerifiedTx{}
	v := ex.store.Get(pb.IBTPKey(id))
	if v != nil {
		if err := verifiedTx.Unmarshal(v); err != nil {
			return nil, false, err
		}
		return verifiedTx.Tx.GetIBTP(), verifiedTx.Valid, nil
	}
	ibtp, isValid, err := ex.syncer.QueryIBTP(id)
	if err != nil {
		if errors.Is(err, syncer.ErrIBTPNotFound) {
			ex.logger.Panicf("query ibtp by id %s from bitxhub: %s", id, err.Error())
		}
		return nil, false, fmt.Errorf("query ibtp from bitxhub: %s", err.Error())
	}

	return ibtp, isValid, nil
}
