package router

import (
	"context"
	"errors"

	"github.com/link33/sidecar/internal/checker"
	"github.com/link33/sidecar/internal/port"
	"github.com/link33/sidecar/internal/repo"
	"github.com/link33/sidecar/model/pb"
	"github.com/sirupsen/logrus"
)

type router struct {
	logger    logrus.FieldLogger
	ctx       context.Context
	cancel    context.CancelFunc
	checker   checker.Checker
	portMap   *port.PortMap
	methodMap map[string]routeMethod
}

type routeMethod func([]string) []port.Port

func NewRouter(logger logrus.FieldLogger) Router {
	ctx, cancel := context.WithCancel(context.Background())
	return &router{
		logger:  logger,
		ctx:     ctx,
		cancel:  cancel,
		portMap: port.NewPortMap(),
	}
}

func (r *router) Start() error {
	r.methodMap["single"] = r.Single
	r.methodMap["multicast"] = r.Multicast
	r.methodMap["broadcast"] = r.Broadcast
	r.methodMap["official"] = r.Official
	return nil
}

func (r *router) Stop() error {
	return nil
}

//TODO
func (r *router) Add(p port.Port) error {
	r.portMap.Add(p)
	go func() {
		c := p.ListenIBTPX()
		for {
			select {
			case ibtpx := <-c:
				err := r.Route(ibtpx)
				if err != nil {
					r.logger.Error(err)
				}
			}
		}
	}()

	return nil
}

func (r *router) Adds(p []port.Port) error {
	for _, pt := range p {
		r.Add(pt)
	}
	return nil
}

func (r *router) Load(key string) (value port.Port, ok bool) {
	return r.portMap.Load(key)
}
func (r *router) Remove(p port.Port) error {
	r.portMap.Remove(p)
	return nil
}

func (r *router) Route(ibtpx *pb.IBTPX) error {
	mode := ibtpx.Mode
	//本网关已签名、中继链已背书、to是本网关内部的appchain，即顺利通过并转发，否则打断。
	if !((r.isSign(ibtpx) && mode == repo.RelayMode && r.isEndorse(ibtpx)) || !r.isSign(ibtpx)) {
		return nil
	}
	//本网关签名
	if !r.isSign(ibtpx) {
		r.sign(ibtpx)
	}

	ibtp := ibtpx.Ibtp
	err := r.checker.Check(ibtp)
	if err != nil {
		r.logger.Error("check ibtp: %w", err)
		return err
	}
	_, to := ibtp.From, ibtp.To
	if pp, is := r.portMap.Port(to); is {
		switch {
		case pp.Type() == port.Sidecar:
			return pp.AsyncSend(ibtpx) //转发给sidecar节点，或者本身local节点。
		case pp.Type() == port.Hub: //发给hub appchain
			return pp.AsyncSend(ibtpx)
		case pp.Type() == port.Appchain: //TODO 本机找到的appchain只能是自己的appchain
			switch mode {
			case repo.RelayMode:
				hub, is := r.getHub()
				if is && !r.isEndorse(ibtpx) {
					return hub.AsyncSend(ibtpx)
				} else {
					return pp.AsyncSend(ibtpx)
				}
			case repo.DirectMode:
				return pp.AsyncSend(ibtpx)
			default:
				//TODO  跳过
				return nil
			}
		default:
			//TODO 跳过
			return nil
		}
	}
	//规则判断 转发给sidecar节点
	method := ibtpx.RouteMethod
	if md, is := r.methodMap[method]; is {
		ports := md(ibtpx.RouteMethodArg)
		if len(ports) == 0 {
			r.firstRoute(ibtpx)
		}
		for _, p := range ports {
			_ = p.AsyncSend(ibtpx)
		}
	} else {
		r.firstRoute(ibtpx)
	}
	return nil
}

func (r *router) firstRoute(ibtp *pb.IBTPX) {
	panic("implement me")
}

func (r *router) getHub() (port.Port, bool) {

	return nil, false
}

func (r *router) isSign(ibtpx *pb.IBTPX) bool {
	panic("implement me")
}

func (r *router) sign(ibtpx *pb.IBTPX) {
	panic("implement me")
}

// hub endorse
func (r *router) isEndorse(ibtpx *pb.IBTPX) bool {
	panic("implement me")
}

func (r *router) HandlerMethod() {}

func (r *router) Single([]string) []port.Port {
	panic("implement me")
}

func (r *router) Multicast([]string) []port.Port {

	panic("implement me")
}

func (r *router) Broadcast([]string) []port.Port {
	panic("implement me")
}

func (r *router) Official([]string) []port.Port {
	panic("implement me")
}

func (r *router) Send(id string, msg port.Message) (*pb.Message, error) {
	if p, is := r.portMap.Port(id); is {
		return p.Send(msg)
	}

	return nil, errors.New("id error!")
}

func (r *router) AsyncSend(id string, msg port.Message) error {
	if p, is := r.portMap.Port(id); is {
		return p.AsyncSend(msg)
	}

	return errors.New("id error!")
}

func (r *router) InPut(ibtp *pb.IBTP) chan *pb.IBTP {
	panic("implement me")
}

func (r *router) OutPut(ibtp *pb.IBTP) chan *pb.IBTP {
	panic("implement me")
}

type RouteMethod interface {
	Single()
	Multicast()
	Broadcast()
	Official()
}
